'''

Jobs for applying textual edits to the LSP client

'''

from typing import List, Union
import pygls
from pygls.server import LanguageServer
from lsprotocol.types import (
    ApplyWorkspaceEditParams,
    Range,
    Position,
)

from threading import Thread
from queue import Queue, Empty
import time
from dataclasses import dataclass
from uniteai.common import find_tag, workspace_edit, workspace_edits, mk_logger
import concurrent
import logging

log = mk_logger('edit', logging.WARN)


##################################################
# Types


@dataclass
class BlockJob:
    uri: str
    start_tag: str
    end_tag: str
    text: str
    # strict jobs MUST be applied, non-strict may be skipped (eg interim state
    # when streaming)
    strict: bool


@dataclass
class DeleteJob:
    uri: str
    regexs: List[str]
    strict: bool


@dataclass
class InsertJob:
    uri: str
    text: str
    line: int
    column: int
    strict: bool


LSPJob = Union[InsertJob, BlockJob, DeleteJob]


##################################################
# Util

def drain_non_strict_queue(q):
    '''Drain a queue up until the first "strict" job or the latest job.'''
    x = None
    while True:
        try:
            x = q.get(False)
            if x.strict:
                return x
            else:
                continue
        except Empty:
            return x


##################################################
#

class Edits:
    '''Edits are saved as jobs to the queue, and applied when document
    versioning aligns properly. Each key is a separate queue. Each queue
    maintains an expectation that only the most recent edit matters, and
    previous edits in the queue have been deprecated, and therefore can be
    dropped.

    '''

    def __init__(self, applicator_fn, job_delay=0.2):
        # A function for applying jobs
        self.applicator_fn = applicator_fn
        self.job_delay = job_delay
        # A dict of queues, one for each separate process of edit blocks.
        self.edit_jobs = {}
        self.job_thread = None

    def create_job_queue(self, name: str):
        ''' Create a job queue. '''
        if name not in self.edit_jobs:
            q = Queue()
            self.edit_jobs[name] = q
            return q
        return self.edit_jobs[name]

    def add_job(self, name: str, job):
        ''' Add a "job", which can be any dataclass and have any data, but must
        also have a field called `strict` which determines if this edit must be
        applied, or can possibly be dropped (eg in the case of streaming data,
        interim state can be dropped.  '''
        log.debug(
            f'ADDING JOB, current queue size: {self.edit_jobs[name].qsize()}'
        )
        self.edit_jobs[name].put(job)

    def start(self):
        self.job_thread = Thread(target=self._process_edit_jobs, daemon=True)
        self.job_thread.start()

    def _process_edit_jobs(self):
        '''A thread that continuously pulls jobs from the queue and attempts to
        execute them.

        Rules:

        * If a non-strict job comes in, try to apply it, until a newer job
          comes in, in which case this job is ok to drop. This is useful during
          streaming.

        * If a strict job comes in, it must be applied before drawing more jobs
          off the queue.

        '''
        n_retries = 5
        n_retries_strict = 5
        failed_job = None
        failed_count = 0
        while True:
            for k, q in self.edit_jobs.items():
                try:
                    job = None  # reset

                    # Strict jobs must apply before continuing to pull off the
                    # queue.
                    if failed_job and failed_job.strict:
                        if failed_count > n_retries_strict:
                            msg = (
                                f'a strict job has failed to apply after '
                                f'{n_retries_strict} attempts. The job was: '
                                f'{failed_job}')
                            logging.error(msg)
                            failed_job = None
                            failed_count = 0
                        else:
                            job = failed_job

                    # If no strict failed_jobs exist, try getting latest off queue,
                    # or previous non-strict failed jobs.
                    else:
                        # get next strict, or the latest (dropping interim)
                        job = drain_non_strict_queue(q)
                        if job:
                            failed_job = None
                            failed_count = 0

                        # retry failed jobs, if no new tasks exist
                        if not job and failed_count <= n_retries:
                            job = failed_job

                    # Execute extant jobs.
                    if job is not None:
                        success = self.applicator_fn(job)
                        if success:
                            failed_job = None
                            failed_count = 0
                        if not success:
                            failed_job = job
                            failed_count += 1

                except concurrent.futures.CancelledError:
                    log.error('Critical Error, process_edit_jobs told to stop.')

                except Exception as e:
                    log.error(f"Exception type: {type(e)}")
                    log.error(f"Exception args: {e.args}")
                    safe_message = str(e).encode('utf-8', errors='replace').decode('utf-8')
                    log.error(f'PROCESS_EDIT_JOBS FAILED: {safe_message}')

            time.sleep(self.job_delay)


def _attempt_edit_job(ls: LanguageServer, job: LSPJob):
    if isinstance(job, InsertJob):
        return _attempt_insert_job(ls, job)
    if isinstance(job, BlockJob):
        return _attempt_block_job(ls, job)
    elif isinstance(job, DeleteJob):
        return _attempt_delete_job(ls, job)


def _attempt_insert_job(ls: LanguageServer, job: InsertJob):
    ''' An `applicator_fn` specialized to pygls Servers.

    Try to execute a job that inserts some text. May fail if document versions
    don't match.

    '''
    log.debug('APPLYING INSERT')
    try:
        doc = ls.workspace.get_document(job.uri)
        if not doc:
            log.error(f'Document not managed by Workspace. Make sure this file type is managed by the client, so it sends `didOpen`. uri={job.uri}')
            return False

        version = doc.version
        log.debug(f'INSERT: uri={job.uri}, doc={doc}, version={version}')
        position = Position(job.line, job.column)
        edit = workspace_edit(job.uri,
                              version,
                              position,
                              position,
                              job.text)
        params = ApplyWorkspaceEditParams(edit=edit)
        future = ls.lsp.send_request("workspace/applyEdit", params)
        future.result()  # blocks
        return True

    except pygls.exceptions.JsonRpcException as e:
        # Most likely a document version mismatch, which is fine. It just
        # means someone edited the document concurrently, and this is set up
        # to try applying the job again.
        log.info(f'ATTEMPT_INSERT_JOB: {e}')
        return False


def _attempt_delete_job(ls: LanguageServer, job: DeleteJob):
    ''' An `applicator_fn` specialized to pygls Servers.

    Try to execute a job that deletes some regex. May fail if document versions
    don't match.

    '''
    try:
        doc = ls.workspace.get_document(job.uri)
        if not doc:
            log.error(f'Document not managed by Workspace. Make sure this file type is managed by the client, so it sends `didOpen`. uri={job.uri}')
            return False
        version = doc.version
        doc_lines = doc.source.split('\n')

        # Collect positions of regexs
        failed_regexs = []
        start_end_texts = []
        for regex in job.regexs:
            m_found = find_tag(regex, doc_lines)
            if m_found:
                ix, s, e = m_found
                start_position = Position(ix, s)
                end_position = Position(ix, e)
                start_end_texts.append((start_position, end_position, ''))
            else:
                failed_regexs.append(regex)

        # log warning if tags missed
        if failed_regexs:
            msg = (f'tags not found in document to apply deletion: '
                   f'{failed_regexs}')
            logging.warn(msg)

        # fire off a delete request for the regexs that were found.
        if start_end_texts:
            edit = workspace_edits(job.uri,
                                   version,
                                   start_end_texts)
            params = ApplyWorkspaceEditParams(edit=edit)
            future = ls.lsp.send_request("workspace/applyEdit", params)
            future.result()  # blocks
            return True

    except pygls.exceptions.JsonRpcException as e:
        # Most likely a document version mismatch, which is fine. It just
        # means someone edited the document concurrently, and this is set up
        # to try applying the job again.
        log.info(f'ATTEMPT_DELETE_JOB: {e}')
        return False


def _attempt_block_job(ls: LanguageServer, job: BlockJob):
    ''' An `applicator_fn` specialized to pygls Servers.

    Try to execute a job (apply an edit to the document) within the confines of
    a "block" demarcated by start and end tags. May fail if document versions
    don't match.

    '''
    log.debug('APPLYING BLOCK')
    try:
        doc = ls.workspace.get_document(job.uri)
        if not doc:
            log.error(f'Document not managed by Workspace. Make sure this file type is managed by the client, so it sends `didOpen`. uri={job.uri}. All docs: {ls.workspace.documents.keys()}')
            return False
        version = doc.version
        doc_lines = doc.source.split('\n')
        log.debug(f'BLOCK CONTEXT: uri={job.uri}, version={version}, doc={type(doc)}|{doc}, doc_source[:100]={doc.source[:100]}')

        m_start = find_tag(job.start_tag, doc_lines)
        m_end = find_tag(job.end_tag, doc_lines)
        if m_start and m_end:
            log.debug('BLOCK: found tags, applying edit')
            ix, s, e = m_start
            start_position = Position(ix, e)
            ix, s, e = m_end
            end_position = Position(ix, s)

            edit = workspace_edit(job.uri,
                                  version,
                                  start_position,
                                  end_position,
                                  job.text)
            params = ApplyWorkspaceEditParams(edit=edit)
            future = ls.lsp.send_request("workspace/applyEdit", params)
            future.result()  # blocks
            return True
        elif job.strict:  # couldn't find tags
            log.debug('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n'
                      f'DOCUMENT: {doc.source}\n'
                      '%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%')
            msg = (
                f'tags not found in document {job.uri} to apply edit: '
                f'{job.text}. Tags: {job.start_tag} and {job.end_tag}'
            )
            logging.warning(msg)

    except pygls.exceptions.JsonRpcException as e:
        # Most likely a document version mismatch, which is fine. It just
        # means someone edited the document concurrently, and this is set up
        # to try applying the job again.
        log.info(f'ATTEMPT_BLOCK_JOB: {e}')
        return False


##################################################
# Smart constructors

def init_block(edit_name,
               tags,
               uri,
               range_or_pos: Union[Range, Position],
               edits):
    ''' Insert new tags, demarcating a new block at the end of the highlighted
    range. '''
    if isinstance(range_or_pos, Range):
        range = range_or_pos
    elif isinstance(range_or_pos, Position):
        range = Range(start=range_or_pos, end=range_or_pos)

    tags = '\n'.join(tags)
    job = InsertJob(
        uri=uri,
        text=tags,
        line=range.end.line+1,
        column=0,
        strict=True,
    )
    edits.add_job(edit_name, job)


def cleanup_block(edit_name, tags, uri, edits):
    ''' Delete tags that demarcated a block. '''
    edits.add_job(edit_name, DeleteJob(
        uri=uri,
        regexs=tags,
        strict=True,
    ))



'''

An opinionated text-embedding suite for use with `document_chat`.

TODO: This should probably use `pgvector` eventually


DEPENDENCIES:
- InstructorEmbedding
- sentence-transformers

Note, I'm not sure which embedding lib I like most; need to test.

'''

from scipy.signal import savgol_filter
from sklearn.metrics.pairwise import cosine_similarity
from typing import List
import logging
import numpy as np
import os
from uniteai.common import mk_logger, read_unicode
import hashlib
import sys
from typing import Union
import sqlite3

# TQDM can be useful for debugging, but prints to stderr which makes VSCode
# flip out.
DEBUG = False
if DEBUG:
    from tqdm import tqdm
else:
    def tqdm(x, *args, **kwargs):
        return x

log = mk_logger('document_chat', logging.DEBUG)


##################################################

def generate_embeddings(model,
                        document: str,
                        window_size: int,
                        stride: int) -> List[np.ndarray]:
    embeddings = []
    offsets = list(enumerate(range(0, len(document) - window_size + 1, stride)))
    for emb_i, doc_i in tqdm(offsets):
        chunk = document[doc_i:doc_i+window_size]
        chunk_e = model.encode(chunk)
        embeddings.append(chunk_e)
    return embeddings


def segments(similarities, document, threshold=0.0):
    out = ''
    last_thresh = False  # for finding edge

    text = ''
    sims = []
    out = []  # [(text, sims), ...]
    for sim, char in zip(similarities, document):
        super_thresh = sim > threshold
        # no longer a super_thresh run
        if last_thresh and not super_thresh:
            out.append((text, np.array(sims)))
            text = ''
            sims = []

        # is a super_thresh run
        if super_thresh:
            text += char
            sims.append(sim)
        last_thresh = super_thresh
    if len(text) > 0:
        out.append((text, np.array(sims)))

    return out


def rank(segments, rank_fn):
    '''Sort segments according to an aggregate function of their scores.'''
    scores = []
    for text, sims in segments:
        scores.append(rank_fn(sims))
    out = []
    for score, (text, sims) in reversed(sorted(zip(scores, segments))):
        out.append(text)
    return out


class Embedding:
    '''
    Embeddings helper that handles things like caching embeddings
    '''

    def __init__(self, model, cache_dir: str):
        self.model = model
        self.cache_dir = cache_dir
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

    def _cache_path(self, path: str, window_size: int, stride: int) -> str:
        name = hashlib.sha1(path.encode()).hexdigest()
        return os.path.join(self.cache_dir, f"{name}_{window_size}_{stride}.npy")

    def _load_embeddings_from_cache(self, path: str, window_size: int, stride: int) -> List[np.ndarray]:
        cache_path = self._cache_path(path, window_size, stride)
        if os.path.exists(cache_path):
            return np.load(cache_path, allow_pickle=True)
        return None

    def _save_embeddings_to_cache(self, path: str, window_size: int, stride: int, embeddings: List[np.ndarray]):
        cache_path = self._cache_path(path, window_size, stride)
        np.save(cache_path, embeddings)

    def embeddings(self, path: str, window_size: int, stride: int) -> List[np.ndarray]:
        embeddings = self._load_embeddings_from_cache(path, window_size, stride)
        if embeddings is None:
            document = read_unicode(path)
            embeddings = generate_embeddings(self.model, document, window_size, stride)
            self._save_embeddings_to_cache(path, window_size, stride, embeddings)
        return embeddings


class Search:
    def __init__(self, model, cache_dir: str,
                 embedding: Embedding,
                 percentile: int = 75,
                 denoise_window_size: int = 1000,
                 denoise_poly_order: int = 2):
        self.model = model
        self.embedding = embedding
        self.percentile = percentile
        self.denoise_window_size = denoise_window_size
        self.denoise_poly_order = denoise_poly_order
        self.cache_dir = cache_dir

    def _similar_tokens(self,
                        query: str,
                        embeddings: List[np.ndarray],
                        window_size: int,
                        stride: int) -> List[float]:
        '''
        Compare a `query` to a strided window over a `document`.
        '''
        document_length = len(embeddings) * stride + window_size - 1
        similarities = np.zeros(document_length, dtype=float)
        overlaps = np.zeros(document_length, dtype=float)
        # VScode interprets progress bar (stderr) as error
        query_e = self.model.encode(query, show_progress_bar=False)
        offsets = list(range(0, document_length - window_size + 1, stride))
        for chunk_e, doc_i in tqdm(zip(embeddings, offsets)):
            sim = cosine_similarity(query_e.reshape(1, -1),
                                    chunk_e.reshape(1, -1))[0][0]
            for j in range(doc_i, doc_i + window_size):
                similarities[j] += sim
                overlaps[j] += 1

        similarities /= np.where(overlaps != 0, overlaps, 1)
        return similarities

    def _denoise_similarities(self, similarities):
        '''
        Apply Savitzky-Golay filter to smooth out the similarity scores.
        '''
        # return similarities

        if len(similarities) < self.denoise_window_size:
            return similarities
        else:
            return savgol_filter(similarities,
                                 self.denoise_window_size,
                                 self.denoise_poly_order)

    def _tune_percentile(self, xs):
        '''
        0-out all elements below percentile. Essentially, this will leave some
        `1-percentile` percentage of the document highlighted.
        '''
        xs = np.copy(xs)
        p = np.percentile(xs, self.percentile)
        xs[xs < p] *= 0
        return xs

    def search(self,
               path: str,
               query: str,
               window_size: int,
               stride: int,
               top_n=5,
               visualize=False):
        '''
        Entry point for the search operation.
        '''
        document = read_unicode(path)
        embeddings = self.embedding.embeddings(path, window_size, stride)
        similarities = self._similar_tokens(query, embeddings, window_size, stride)
        # # Instruct Embeddings needs this
        # last_edge = int(len(similarities) * 0.01)
        # similarities[-last_edge:] = similarities[-last_edge]
        d_similarities = self._denoise_similarities(similarities)
        d_similarities -= d_similarities.min()
        d_similarities /= d_similarities.max()
        d_similarities = self._tune_percentile(d_similarities)

        segs = segments(d_similarities, document)
        # ranked_segments = rank(segs, np.mean)[:top_n]
        ranked_segments = rank(segs, np.max)[:top_n]
        if visualize:
            import matplotlib.pyplot as plt
            plt.plot(similarities)
            plt.plot(d_similarities)
            plt.show()
        return ranked_segments


##################################################
# Dev API

def load_from_database(connection, title: Union[str, None], url):
    cursor = connection.cursor()
    if title is not None:
        cursor.execute("""
SELECT title, file_path, content, url FROM resources
WHERE title = ? OR url = ?
""", (title, url))
    else:
        cursor.execute("""
SELECT title, file_path, content, url FROM resources
WHERE url = ?
""", (url,))
    result = cursor.fetchone()
    return result if result else (None, None, None, None)


def search_helper(search,
                  db_path,
                  titles_urls,
                  query,
                  window_size=200, stride=50, top_n=5, visualize=False):
    with sqlite3.connect(db_path) as conn:
        results = []
        for title, url in titles_urls:
            title, file_path, content, url = load_from_database(conn, title, url)
            if not title:
                continue
            result = search.search(file_path, query, window_size, stride, top_n, visualize)
            results.append((url, title, result))
        return results



'''

Text-to-Speech

pip install spacy
python -m spacy download en_core_web_sm

'''

import re
import os
from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Range,
    TextDocumentIdentifier,
    WorkspaceEdit,
)
from threading import Event, Thread
from thespian.actors import Actor
import argparse
import logging
import importlib

from uniteai.common import extract_range, mk_logger

os.environ["SUNO_USE_SMALL_MODELS"] = "True"  # must do before `import bark`


# Temperatures
#   1.0 = more diverse
#   0.0 = more conservative
TEXT_TEMP = 0.6
WAVEFORM_TEMP = 0.8

# START_TAG = ':START_DOCUMENT_CHAT:'
# END_TAG = ':END_DOCUMENT_CHAT:'
NAME = 'text_to_speech'

# A custom logger for just this feature. You can tune the log level to turn
# on/off just this feature's logs.
log = mk_logger(NAME, logging.DEBUG)


def prep_prompts(text, desired_length=200, max_length=300):
    """Split text it into chunks of a desired length trying to keep sentences
    intact.

    From tortoise-tts"""
    # normalize text, remove redundant whitespace and convert non-ascii quotes
    # to ascii
    text = re.sub(r'\n\n+', '\n', text)
    text = re.sub(r'\s+', ' ', text)
    text = re.sub(r'[“”]', '"', text)

    rv = []
    in_quote = False
    current = ""
    split_pos = []
    pos = -1
    end_pos = len(text) - 1

    def seek(delta):
        nonlocal pos, in_quote, current
        is_neg = delta < 0
        for _ in range(abs(delta)):
            if is_neg:
                pos -= 1
                current = current[:-1]
            else:
                pos += 1
                current += text[pos]
            if text[pos] == '"':
                in_quote = not in_quote
        return text[pos]

    def peek(delta):
        p = pos + delta
        return text[p] if p < end_pos and p >= 0 else ""

    def commit():
        nonlocal rv, current, split_pos
        rv.append(current)
        current = ""
        split_pos = []

    while pos < end_pos:
        c = seek(1)
        # do we need to force a split?
        if len(current) >= max_length:
            if len(split_pos) > 0 and len(current) > (desired_length / 2):
                # we have at least one sentence and we are over half the
                # desired length, seek back to the last split
                d = pos - split_pos[-1]
                seek(-d)
            else:
                # no full sentences, seek back until we are not in the middle
                # of a word and split there
                while c not in '!?.\n ' and pos > 0 and len(current) > desired_length:
                    c = seek(-1)
            commit()
        # check for sentence boundaries
        elif not in_quote and (c in '!?\n' or (c == '.' and peek(1) in '\n ')):
            # seek forward if we have consecutive boundary markers but still
            # within the max length
            while pos < len(text) - 1 and len(current) < max_length and peek(1) in '!?.':
                c = seek(1)
            split_pos.append(pos)
            if len(current) >= desired_length:
                commit()
        # treat end of quote as a boundary if its followed by a space or newline
        elif in_quote and peek(1) == '"' and peek(2) in '\n ':
            seek(2)
            split_pos.append(pos)
    rv.append(current)

    # clean up, remove lines with only whitespace or punctuation
    rv = [s.strip() for s in rv]
    rv = [s for s in rv if len(s) > 0 and not re.match(r'^[\s\.,;:!?]*$', s)]

    return rv

# def prep_prompts(text_prompt: str):
#     ''' Bark prompts should ideally be around 14seconds once translated. Here's
#     a heuristic for that. '''

#     # split into sentences (nltk)
#     sentences = nltk.sent_tokenize(text_prompt)  # import separately since it's slow to load

#     # 2 sentences = 14 seconds
#     return [sentences[i] + " " + sentences[i + 1]
#             for i in range(0, len(sentences), 2)]


class TextToSpeechActor(Actor):
    def __init__(self):
        self.should_stop = Event()
        self.is_running = False
        self.audio_history = None  # bark can mimic this history

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        doc = msg.get('doc')
        edits = msg.get('edits')

        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
locked: {self.should_stop.is_set()}
%%%%%%%%%%''')

        ##########
        #
        if command == 'save':
            prompt = msg.get('prompt')
            self.save(prompt)

        ##########
        #
        if command == 'play':
            prompt = msg.get('prompt')
            self.play(prompt)

        ##########
        # Stop
        elif command == 'stop':
            self.stop()

        ##########
        # Set Config
        elif command == 'initialize':
            log.info('Warming up TTS model in separate thread')
            warmup_thread = Thread(target=self._warmup)
            warmup_thread.daemon = True
            warmup_thread.start()

    def _warmup(self):
        ''' Warm up, intended for a separate thread. '''
        # global nltk
        global bark
        global sounddevice

        # nltk = importlib.import_module('nltk')  # loads slowly
        # nltk.download('punkt')

        bark = importlib.import_module('bark')
        sounddevice = importlib.import_module('sounddevice')

        bark.preload_models()
        log.info('Warmed up.')

    def save(self, prompt):
        prompt = prompt.strip()
        if re.findall(r'.+_speaker_.+', prompt):
            self.audio_history = prompt
            return True

        log.debug(f'SAVE: {str(self.audio_history)[:100]}')
        self.should_stop.clear()

        history, audio_array = bark.generate_audio(
            prompt,
            text_temp=TEXT_TEMP,
            waveform_temp=WAVEFORM_TEMP,
            output_full=True,
            silent=True  # no progress bar
        )
        self.audio_history = history
        # self.audio_history = prompt
        sounddevice.play(audio_array, samplerate=bark.SAMPLE_RATE)

    def play(self, prompt):
        log.debug(f'PLAY: {str(self.audio_history)[:100]}')
        self.should_stop.clear()
        sentences = prep_prompts(prompt)
        for sentence in sentences:
            # break early
            if self.should_stop.is_set():
                return False
            audio_array = bark.generate_audio(
                sentence.strip(),
                history_prompt=self.audio_history,
                text_temp=TEXT_TEMP,
                waveform_temp=WAVEFORM_TEMP,
                silent=True  # no progress bar
            )
            # break early
            if self.should_stop.is_set():
                return False
            sounddevice.wait()  # wait for previous gen
            sounddevice.play(audio_array, samplerate=bark.SAMPLE_RATE)

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running:
            log.info('WARN: ON_STOP_BUT_STOPPED')
        self.should_stop.set()
        log.debug('FINALLY STOPPED')


def code_action_text_to_speech_save(params: CodeActionParams):
    ''' A code_action triggers a command, which sends a message to the Actor,
    to handle it. '''
    text_document = params.text_document
    # position of the highlighted region in the client's editor
    range = params.range
    return CodeAction(
        title='TextToSpeech Save History',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='TextToSpeech Save History',
            command='command.textToSpeechSave',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, range]
        )
    )

def code_action_text_to_speech_play(params: CodeActionParams):
    ''' A code_action triggers a command, which sends a message to the Actor,
    to handle it. '''
    text_document = params.text_document
    # position of the highlighted region in the client's editor
    range = params.range
    return CodeAction(
        title='TextToSpeech Play',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='TextToSpeech Play',
            command='command.textToSpeechPlay',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, range]
        )
    )


##################################################
# Setup
#
# NOTE: In `.uniteai.yml`, just add `uniteai.text_to_speech` under `modules`,
#       and this will automatically get built into the server at runtime.
#

def configure(config_yaml):
    parser = argparse.ArgumentParser()
    args, _ = parser.parse_known_args()
    return args


def initialize(config, server):

    # Actor
    server.add_actor(NAME, TextToSpeechActor)

    # Initialize configuration in Actor
    server.tell_actor(NAME, {
        'command': 'initialize',
        'config': config,
    })

    # CodeActions
    server.add_code_action(code_action_text_to_speech_save)
    server.add_code_action(code_action_text_to_speech_play)

    # Modify Server
    @server.thread()
    @server.command('command.textToSpeechSave')
    def text_to_speech_save(ls, args):
        if len(args) != 2:
            log.error(f'command.textToSpeechSave: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        range = ls.converter.structure(args[1], Range)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Extract the highlighted region
        prompt = extract_range(doc_source, range)

        # Send a message to start the stream
        actor_args = {
            'command': 'save',
            'uri': uri,
            'prompt': prompt,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)

        # Return null-edit immediately (the rest will stream)
        return WorkspaceEdit()

    # Modify Server
    @server.thread()
    @server.command('command.textToSpeechPlay')
    def text_to_speech_play(ls, args):
        if len(args) != 2:
            log.error(f'command.textToSpeechPlay: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        range = ls.converter.structure(args[1], Range)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Extract the highlighted region
        prompt = extract_range(doc_source, range)

        # Send a message to start the stream
        actor_args = {
            'command': 'play',
            'uri': uri,
            'prompt': prompt,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)

        # Return null-edit immediately (the rest will stream)
        return WorkspaceEdit()



'''

An LSP server that connects to the LLM server and APIs for doing the brainy
stuff.

'''

import logging
import uniteai.config as config
from typing import List, Callable
from pygls.server import LanguageServer
from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    TextDocumentIdentifier,
    TEXT_DOCUMENT_DID_OPEN,
    INITIALIZED,
    WINDOW_SHOW_MESSAGE_REQUEST,
    DidOpenTextDocumentParams,
)
from pygls.protocol import default_converter
from concurrent.futures import ThreadPoolExecutor
from thespian.actors import ActorSystem
import uniteai.edit as edit
from uniteai.common import mk_logger
import importlib
import time
import asyncio
from functools import partial
import sys

NAME = 'lsp_server'
log = mk_logger(NAME, logging.INFO)


##################################################
# Actor Concurrency Model
#
# NOTE: If you try to load a CUDA model on a separate process you get this
#       error, and I haven't debugged yet:
#
#       RuntimeError: Cannot re-initialize CUDA in forked subprocess. To use
#       CUDA with multiprocessing, you must use the 'spawn' start method

CONCURRENCY_MODEL = 'simpleSystemBase'  # multi-threading (easier debug)

# CONCURRENCY_MODEL = 'multiprocQueueBase'  # multi-processing (faster)


##################################################
# Logging
#
# Change the basicConfig level to be the generous DEBUG, quiet the libraries,
# and allow custom loggers in modules to have their own debug levels. This
# helps especially with debugging weird concurrency quirks, ie allowing
# different processes to report things as needed, in a way that can be easily
# turned on and off per object of your attention.
#
# NOTE: Thespian has a domineering logging methodology. To customize their
#       formatter, see: `thespian.system.simpleSystemBase`.
#
#       Also, https://github.com/thespianpy/Thespian/issues/73
#
# NOTE: VSCode thinks the server is erroring if it logs to stdout from the main
#       thread, but spawned threads don't seem to have the same effect.

logging.basicConfig(
    filename='log_file.log',  # TODO: feature loggers still seem to report on stdout
    # stream=sys.stdout,
    level=logging.DEBUG,
)

# Quiet the libs a little
logging.getLogger('pygls.feature_manager').setLevel(logging.WARN)
logging.getLogger('pygls.protocol').setLevel(logging.WARN)
logging.getLogger('Thespian').setLevel(logging.WARN)
logging.getLogger('asyncio').setLevel(logging.WARN)


##################################################
# Module Loading
#
# All features on opt-in only, and this loads them up (based on .uniteai.yml
# config)

def load_module(module_name, config_yaml, server):
    ''' Load individual feature modules. This is useful for loading just the
    modules referenced in the config.

    It loads in 3 steps:

    1. collect config

    2. fast initialization (before connecting to client)

    3. post initialization (after connecting to client) '''
    logging.info(f'Loading module: {module_name}')

    # COLLECT CONFIGURATION (eg yml + cli args)
    module = importlib.import_module(module_name)
    if hasattr(module, 'configure'):
        logging.info(f'Configuring module: {module_name}')
        args = module.configure(config_yaml)
    else:
        logging.warn(f'No `configure` fn found for: {module_name}')

    # INITIALIZE. This happens before client connection.
    if hasattr(module, 'initialize'):
        module.initialize(args, server)
        logging.info(f'Initializing module: {module_name}')
    else:
        logging.warn(f'No `initialize` fn found for: {module_name}')

    # POST-INITIALIZATION. These will be called post client conection.
    if hasattr(module, 'post_initialization'):
        server.post_initialization.append(
            partial(module.post_initialization, args)
        )
    else:
        logging.warn(f'No `post_initialization` fn found for: {module_name}')


##################################################
# Server

class Server(LanguageServer):
    def __init__(self, name, version):
        super().__init__(name, version)
        self.code_actions = []

        # Post client-connection. Functions in this list will be called. This is
        # useful for resource-intensive setup that shouldn't block the client
        # booting up.
        self.post_initialization = []

        # A ThreadPool for tasks. By throwing long running tasks on here, LSP
        # endpoints can return immediately.
        self.executor = ThreadPoolExecutor(max_workers=3)

        # For converting jsonified things into proper instances. This is
        # necessary when a CodeAction summons a Command, and passes raw JSON to
        # it.
        self.converter = default_converter()

        # Edit-jobs
        edit_job_delay = 0.2
        self.edits = edit.Edits(
            lambda job: edit._attempt_edit_job(self, job), edit_job_delay)
        self.edits.start()

        # Actor System
        self.actor_system = ActorSystem(CONCURRENCY_MODEL)
        self.actors = {}  # each feature (eg Local LLM) gets its own actor)

    def load_modules(self, args, config_yaml):
        for module_name in args.modules:
            start_time = time.time()  # keep track of initialization time
            load_module(module_name, config_yaml, self)
            end_time = time.time()
            log.info(f'Loading module {module_name} took {end_time - start_time:.2f} seconds.')

    def add_actor(self, name, cls):
        ''' Adds an Actor, and an Edits job queue for that Actor. '''
        self.edits.create_job_queue(name)
        actor = self.actor_system.createActor(
            cls,
            globalName=name
        )
        self.actors[name] = actor
        return actor

    def add_code_action(self, f: Callable[[CodeActionParams], CodeAction]):
        ''' Code Actions are an LSP term, and your client will show you
        available Actions. '''
        self.code_actions.append(f)

    def tell_actor(self, name, msg):
        ''' A non-blocking signal to control an Actor. '''
        self.actor_system.tell(self.actors[name], msg)


def create_action(title, cmd):
    ''' A Helper for adding Commands as CodeActions. '''
    def code_action_fn(params: CodeActionParams):
        text_document = params.text_document
        range = params.range
        return CodeAction(
                    title=title,
                    kind=CodeActionKind.Refactor,
                    command=Command(
                        title=title,
                        command=cmd,
                        arguments=[text_document, range]
                    )
                )
    return code_action_fn


def initialize(args, config_yaml):
    '''
    A Barebones pygls LSP Server.
    '''
    server = Server("uniteai", "0.0.0")

    @server.feature(INITIALIZED)
    def on_initialized(ls, initialized_params):
        log.info('Initialized. Calling `post_initialization` functions.')
        for f in ls.post_initialization:
            f(ls)

    @server.feature('workspace/didChangeConfiguration')
    def workspace_did_change_configuration(ls: Server, args):
        ''' There's a warning without this. '''
        log.debug(f'workspace/didChangeConfiguration: args={args}')
        return []

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    async def did_open(ls: Server, params: DidOpenTextDocumentParams):
        """Text document did open notification. It appears this does not
        overwrite the default `didOpen` handler in pygls, so we can add extra
        `didOpen` logic here."""
        log.debug(f'Document opened: uri={params.text_document.uri}')

    @server.feature("textDocument/codeAction")
    def code_action(params: CodeActionParams) -> List[CodeAction]:
        ''' Code Actions are eg a list of options you can trigger in your
        client. Included modules can add to `server.code_actions` via
        `server.add_code_action`.  '''
        return [action(params) for action in server.code_actions]

    @server.thread()
    @server.command('command.stop')
    def stop(ls: Server, args):
        '''
        Tell ALL actors to stop.
        '''
        if len(args) != 1:
            log.error(f'command.stop: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Send a message to stop the stream
        actor_args = {
            'command': 'stop',
            'uri': uri,
            'edits': ls.edits,
            'doc': doc_source,
        }
        for actor in ls.actors:
            ls.tell_actor(actor, actor_args)

    # Add `command.stop` as a "Code Action" too (accessible from the dropdown
    # menu, eg `M-'`.
    server.add_code_action(create_action('Stop Streaming Things', 'command.stop'))
    return server


##################################################
# Main

def main():
    # First pass at configuration. Further passes will pick up config
    # per-feature.
    args, config_yaml, parser = config.fetch_config()
    server = initialize(args, config_yaml)

    # Server features must be registered to the server before connecting. IE you
    # cannot add `@server.feature(...)` after the server has been
    # started. Capabilities *can* be dynamically registered, but since clients
    # don't need to implement that we should avoid it by just registering ahead
    # of time. This means that you have to be careful to not do
    # resource-intensive things before `server.start_io()` or the boot time will
    # be slow and clients will timeout.
    server.load_modules(args, config_yaml)

    if args.tcp:
        logging.info(f'Starting LSP on port {args.lsp_port}')
        server.start_tcp(host='localhost', port=args.lsp_port)
    elif args.stdio:
        logging.info('Starting on STDIO')
        server.start_io()

if __name__ == '__main__':
    main()



'''

Commonly needed functions. TODO: should probably be relocated

'''

import re
from typing import List, Tuple
from lsprotocol.types import (
    Position,
    Range,
    OptionalVersionedTextDocumentIdentifier,
    TextEdit,
    WorkspaceEdit,
)

from pypdf import PdfReader
from bs4 import BeautifulSoup
from pathlib import Path
from io import BytesIO
import nbformat
import logging


##################################################


def mk_logger(name, level):
    ''' A logger builder helper. This helps out since Thespian is overly
    opinionated about logging.'''
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)s:%(name)s => %(message)s [%(pathname)s:%(lineno)d]'
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


##################################################
# Dict helpers

def get_nested(config, keys):
    ''' Ex: get_nested(my_dict, ['some', 'nested', 'key'] '''
    temp = config
    for k in keys:
        if k in temp:
            temp = temp[k]
        else:
            return None
    return temp


##################################################
# String helpers

def insert_text_at(document, insert_text, line_no, col_no):
    # Split the document into lines
    lines = document.split('\n')

    # Validate line number
    if not (-1 <= line_no <= len(lines)):
        raise ValueError('Line number out of range')
    is_end_line = line_no == len(lines) or line_no == -1
    line = '' if is_end_line else lines[line_no]

    # Validate Column number
    if col_no >= len(line)+1 or col_no < -1:
        raise ValueError('Column number out of range')
    is_end_col = col_no == len(line) or col_no == -1

    if is_end_col:
        updated_line = line + insert_text
    else:
        updated_line = line[:col_no] + insert_text + line[col_no:]

    if is_end_line:
        lines += [updated_line]
    else:
        lines[line_no] = updated_line
    return '\n'.join(lines)


def find_tag(tag: str, doc_lines: [str]):
    ''' Find index of first element that contains `tag`. '''
    ix = 0
    for ix, line in enumerate(doc_lines):
        match = re.search(tag, line)
        if match:
            return ix, match.start(), match.end()
    return None


def find_block(start_tag, end_tag, doc):
    '''Fine the indices of a start/end-tagged block.'''
    doc_lines = doc.split('\n')
    s = find_tag(start_tag, doc_lines)
    e = find_tag(end_tag, doc_lines)
    return s, e


def find_pattern_in_document(
        document: str,
        pattern: str) -> List[Tuple[int, int, int]]:
    '''Return (line, start_col, end_col) for each match. Regex cannot span
    newlines.'''
    result = []
    compiled_pattern = re.compile(pattern)

    for line_number, line in enumerate(document.split('\n')):
        for match in compiled_pattern.finditer(line):
            start, end = match.span()
            result.append((line_number, start, end))

    return result

##################################################
# IO Helpers


def convert_ipynb_to_py(ipynb_buf):
    ''' Convert ipynb to human-readable python source file.'''
    nb = nbformat.read(ipynb_buf, as_version=4)

    # Convert cells to Python code with comments
    python_code = ""
    for cell in nb.cells:
        if cell.cell_type == 'code':
            # Add the code cell's content to the Python code
            python_code += cell.source.strip() + "\n\n"
        elif cell.cell_type == 'markdown':
            # Add the markdown content as comments
            lines = cell.source.strip().split('\n')
            for line in lines:
                python_code += f"# {line}\n"
            python_code += "\n"
    return python_code


def read_unicode(file_path, buf=None):
    '''Read a file_path in as a unicode string.

    Args:
      file_path: a path to a file to read
      buf: if the file has already been read into an IO buffer, we can use it
    '''
    file_path = Path(file_path)

    if buf is None:
        with open(file_path, 'rb') as fp:
            buf = BytesIO(fp.read())

    file_content = buf.getvalue()
    if file_path.suffix == '.pdf':
        pdf = PdfReader(BytesIO(file_content))
        return '\n'.join(page.extract_text() for page in pdf.pages)
    elif file_path.suffix == '.html':
        soup = BeautifulSoup(file_content, "html.parser")
        return soup.get_text()
    elif file_path.suffix == '.txt' or file_path.suffix == '.json':
        return file_content.decode('utf-8')
    elif file_path.suffix == '.txt' or file_path.suffix == '.ipynb':
        py = convert_ipynb_to_py(buf)
        return py
    else:
        return file_content.decode('utf-8', errors='ignore')


##################################################
# LSP-Specific

def workspace_edit(uri: str,
                   version: int,
                   start: Position,
                   end: Position,
                   new_text: str) -> WorkspaceEdit:
    ''' Build a `WorkspaceEdit` for pygls to send to LSP client. '''
    text_edit = TextEdit(range=Range(start=start, end=end), new_text=new_text)
    text_document_edit = {
        'textDocument': OptionalVersionedTextDocumentIdentifier(
            uri=uri,
            version=version
        ),
        'edits': [text_edit],
    }
    return WorkspaceEdit(document_changes=[text_document_edit])


def workspace_edits(uri: str,
                    version: int,
                    start_end_text: List[Tuple[Position, Position, str]]
                    ) -> WorkspaceEdit:
    ''' Build a `WorkspaceEdit` for pygls to send to LSP client. '''
    text_edits = [
        TextEdit(range=Range(start=start, end=end), new_text=new_text)
        for start, end, new_text in start_end_text
    ]
    text_document_edit = {
        'textDocument': OptionalVersionedTextDocumentIdentifier(
            uri=uri,
            version=version
        ),
        'edits': text_edits,
    }
    return WorkspaceEdit(document_changes=[text_document_edit])


def extract_range(doc: str, range: Range) -> str:
    '''Extract the highlighted region of the text coming from the client.'''
    lines = doc.split("\n")
    start = range.start
    end = range.end
    if start.line == end.line:
        return lines[start.line][start.character:end.character]
    else:
        start_line_text = lines[start.line][start.character:]
        middle_lines_text = lines[start.line + 1:end.line]
        end_line_text = lines[end.line][:end.character]
        return "\n".join([start_line_text] +
                         middle_lines_text +
                         [end_line_text])



'''

Start building your feature off this example!

This example inserts auto-incrementing numbers.

'''

import asyncio
import lsprotocol.types as lsp
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from thespian.actors import Actor
import argparse
import logging
import time

from uniteai.edit import init_block, cleanup_block, BlockJob
from uniteai.common import find_block, mk_logger, get_nested

START_TAG = ':START_EXAMPLE:'
END_TAG = ':END_EXAMPLE:'
NAME = 'example'

# A custom logger for just this feature. You can tune the log level to turn
# on/off just this feature's logs.
log = mk_logger(NAME, logging.DEBUG)


class ExampleActor(Actor):
    def __init__(self):
        self.is_running = False
        self.executor = ThreadPoolExecutor(max_workers=3)
        self.current_future = None
        self.should_stop = Event()
        self.tags = [START_TAG, END_TAG]

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        doc = msg.get('doc')
        edits = msg.get('edits')

        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
is_running: {self.is_running}
locked: {self.should_stop.is_set()}
future: {self.current_future}

EDITS STATE:
job_thread alive: {edits.job_thread.is_alive() if edits and edits.job_thread else "NOT STARTED"}
%%%%%%%%%%''')

        ##########
        # Start
        if command == 'start':
            uri = msg.get('uri')
            cursor_pos = msg.get('cursor_pos')
            engine = msg.get('engine')
            max_length = msg.get('max_length')
            edits = msg.get('edits')

            # check if block already exists
            start_ixs, end_ixs = find_block(START_TAG,
                                            END_TAG,
                                            doc)

            if not (start_ixs and end_ixs):
                init_block(NAME, self.tags, uri, cursor_pos, edits)

            self.start(uri, cursor_pos, engine, max_length, edits)

        ##########
        # Stop
        elif command == 'stop':
            self.stop()

        ##########
        # Set Config
        elif command == 'set_config':
            config = msg['config']
            self.start_digit = config.example_start_digit
            self.end_digit = config.example_end_digit
            self.delay = config.example_delay

    def start(self, uri, cursor_pos, engine, max_length, edits):
        if self.is_running:
            log.info('WARN: ON_START_BUT_RUNNING')
            return
        log.debug('ACTOR START')

        self.is_running = True
        self.should_stop.clear()
        self.current_future = self.executor.submit(
            self.stream_fn, uri, self.should_stop, edits
        )
        log.debug('START CAN RETURN')

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running:
            log.info('WARN: ON_STOP_BUT_STOPPED')

        self.should_stop.set()

        if self.current_future:
            self.current_future.result()  # block, wait to finish
            self.current_future = None
        log.debug('FINALLY STOPPED')

    def stream_fn(self, uri, stop_event, edits):
        log.debug('START: EXAMPLE_STREAM_FN')
        try:
            # Stream the results to LSP Client
            running_text = ''
            for x in range(self.start_digit, self.end_digit + 1):
                # For breaking out early
                if stop_event.is_set():
                    log.debug('STREAM_FN received STOP EVENT')
                    break

                running_text += f'{x} '
                job = BlockJob(
                    uri=uri,
                    start_tag=START_TAG,
                    end_tag=END_TAG,
                    text=f'\n{running_text}\n',
                    # "non-strict" means that if this job doesn't get applied
                    # successfully, it can be dropped. This is useful when
                    # streaming.
                    strict=False,
                )
                edits.add_job(NAME, job)
                time.sleep(self.delay)

            # Streaming is done, and those added jobs were all
            # non-strict. Let's make sure to have one final strict
            # job. Streaming jobs are ok to be dropped, but we need to make
            # sure it does finalize, eg before a strict delete-tags job is
            # added.
            job = BlockJob(
                uri=uri,
                start_tag=START_TAG,
                end_tag=END_TAG,
                text=f'\n{running_text}\n',
                strict=True,
            )
            edits.add_job(NAME, job)

        except Exception as e:
            log.error(f'Error: ExapleActor, {e}')

        # Cleanup
        log.debug('CLEANING UP')
        cleanup_block(NAME, self.tags, uri, edits)
        self.is_running = False
        self.current_future = None
        self.should_stop.clear()


def code_action_example(params: lsp.CodeActionParams):
    ''' A code_action triggers a command, which sends a message to the Actor,
    to handle it. '''
    text_document = params.text_document
    # position of the highlighted region in the client's editor
    range = params.range  # lsp spec only provides `Range`
    cursor_pos = range.end
    return lsp.CodeAction(
        title='Example Counter',
        kind=lsp.CodeActionKind.Refactor,
        command=lsp.Command(
            title='Example Counter',
            command='command.exampleCounter',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, cursor_pos]
        )
    )


##################################################
# Setup
#
# NOTE: In `.uniteai.yml`, just add `uniteai.example` under `modules`, and this
#       example feature will automatically get built into the server at
#       runtime.
#

def create_action(title, cmd):
    ''' A Helper for adding Commands as CodeActions. '''
    def code_action_fn(params: lsp.CodeActionParams):
        text_document = params.text_document
        range = params.range
        return lsp.CodeAction(
            title=title,
            kind=lsp.CodeActionKind.Refactor,
            command=lsp.Command(
                title=title,
                command=cmd,
                arguments=[text_document, range]
            )
        )
    return code_action_fn


def configure(config_yaml):
    parser = argparse.ArgumentParser()
    parser.add_argument('--example_start_digit', default=get_nested(config_yaml, ['example', 'start_digit']))
    parser.add_argument('--example_end_digit', default=get_nested(config_yaml, ['example', 'end_digit']))
    parser.add_argument('--example_delay', default=get_nested(config_yaml, ['example', 'delay']))

    # These get picked up as `config` in `initialize`

    # bc this is only concerned with example params, do not error if extra
    # params are sent via cli.
    args, _ = parser.parse_known_args()
    return args


def initialize(config, server):
    # Config
    start_digit = config.example_start_digit
    end_digit = config.example_end_digit
    delay = config.example_delay

    # CodeActions
    server.add_code_action(code_action_example)

    # Modify Server
    @server.thread()
    @server.command('command.exampleCounter')
    def example_counter(ls, args):
        if len(args) != 2:
            log.error(f'command.exampleCounter: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], lsp.TextDocumentIdentifier)
        cursor_pos = ls.converter.structure(args[1], lsp.Position)

        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Send a message to start the stream
        actor_args = {
            'command': 'start',
            'uri': uri,
            'cursor_pos': cursor_pos,
            'start_digit': start_digit,
            'end_digit': end_digit,
            'delay': delay,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)

        # Return null-edit immediately (the rest will stream)
        return lsp.WorkspaceEdit()

    ##############################
    # Example LSP Features

    @server.feature(lsp.TEXT_DOCUMENT_COMPLETION, lsp.CompletionOptions(trigger_characters=['.']))
    async def provide_completions(params: lsp.CompletionParams):
        '''
        NOTES:
        * These are not only triggered on the `trigger_character`. It's just that the `trigger_character` also works.
        '''
        log.info('COMLETIONS STARTED')
        # Extract the line text till the position of the autocomplete trigger
        position = params.position
        uri = params.text_document.uri
        doc = server.workspace.get_document(uri)
        line_text = doc.lines[position.line][:position.character]

        if '.' in line_text:  # If the trigger character was used
            return await triggered_completions()

        return general_completions(line_text)

    def general_completions(starting_text):
        ''' Return general completions based on starting_text. '''
        # Filtering based on existing text
        possible_completions = [
            lsp.CompletionItem(
                label='my_variable_1',
                kind=lsp.CompletionItemKind.Variable,
                detail='A sample variable',
                sort_text='aaa',  # Controlling order
                commit_characters=["."]
            ),
            lsp.CompletionItem(
                label='my_variable_2',
                kind=lsp.CompletionItemKind.Variable,
                detail='A sample variable',
                sort_text='aaa',  # Controlling order
                commit_characters=["."]
            ),
            lsp.CompletionItem(
                label='a whole sentence can be completed on, but the inserted text changed',
                kind=lsp.CompletionItemKind.Variable,
                detail='A long natural language example',
                insert_text='hah tricked you!',
                sort_text='aab'
            ),
            lsp.CompletionItem(
                label='my_function',
                kind=lsp.CompletionItemKind.Function,
                detail='A sample function',
                insert_text='my_function(',  # Auto-insert parentheses
                sort_text='aab'
            ),
            lsp.CompletionItem(
                label='''
def multiline(x):
    a = 123
    b = 456
    c = 789
    return a + b + c
'''.strip(),
                insert_text_format=lsp.InsertTextFormat.PlainText,

                # Yassnippet: a tabbable completion
                # insert_text_format=InsertTextFormat.Snippet,
            ),
            lsp.CompletionItem(
                label='my_class',
                kind=lsp.CompletionItemKind.Class,
                detail='A sample class',
                insert_text='MyClass',

                # This will also add an import statement (or arbitrary edit)
                additional_text_edits=[
                    lsp.TextEdit(
                        lsp.Range(lsp.Position(0, 0), lsp.Position(0, 0)),
                        "import MyClass\n"
                    )
                ],
                sort_text='zzz'
            ),
        ]

        # Filtering completions to only include items that start with starting_text.
        #
        # The client does it's own filtering, but it also does a fuzzy match
        # thing that means if you've typed only a few chars, many of the
        # possible matches *will* match, but probably shouldn't.
        filtered_completions = [item for item in possible_completions if item.label.startswith(starting_text)]

        return lsp.CompletionList(
            # `is_incomplete` means: "I could have given you more completions,
            #     but didn't want to (IE too resource intensive or too many
            #     completions possible. Keep typing to narrow the list."
            is_incomplete=False,
            items=filtered_completions
        )

    async def triggered_completions():
        return lsp.CompletionList(
            is_incomplete=True,
            items=[
                lsp.CompletionItem(
                    label="main",
                    filter_text="mn",  # demonstrate filter_text
                    sort_text='aaa',
                    documentation=None,
                ),
                lsp.CompletionItem(
                    "foo",
                    sort_text='aab',
                    documentation=None,
                ),
                # additional items will be fetched in the next completion call...
            ]
        )

    # Signature Help
    @server.feature(lsp.TEXT_DOCUMENT_SIGNATURE_HELP)
    def provide_signature_help(ls, params):
        ''' The signature help shows up, I think, just in the minibuffer in emacs. '''
        signature = lsp.SignatureInformation(
            label='myFunctionWithSig(arg1: string, arg2: int)',
            documentation='A sample function',
            parameters=[
                lsp.ParameterInformation(label='arg1', documentation='A string argument'),
                lsp.ParameterInformation(label='arg2', documentation='An integer argument'),
            ]
        )
        return lsp.SignatureHelp(signatures=[signature])

    # Goto Definition
    @server.feature(lsp.TEXT_DOCUMENT_DEFINITION)
    def goto_definition(ls, params):
        return [lsp.Location(
            uri='file:///path/to/definition/file.py',
            range=lsp.Range(lsp.Position(line=10, character=5),
                            lsp.Position(line=10, character=15)))
                ]

    # Find References
    @server.feature(lsp.TEXT_DOCUMENT_REFERENCES)
    def find_references(ls, params):
        ''' Perhaps you could use this to get cos-sim of embeddings with a doc,
        and allow you to jump to locations in that doc. '''
        return [
            lsp.Location(uri='file:///path/to/reference1/file.py',
                         range=lsp.Range(
                             lsp.Position(line=5, character=5),
                             lsp.Position(line=5, character=15))),
            lsp.Location(uri='file:///path/to/reference2/file.py',
                         range=lsp.Range(
                             lsp.Position(line=8, character=7),
                             lsp.Position(line=8, character=17)))]

    # Document Highlights
    @server.feature(lsp.TEXT_DOCUMENT_DOCUMENT_HIGHLIGHT)
    def provide_highlights(ls, params):
        ''' NOTE: You must trigger your client to show highlights '''
        return [
            lsp.DocumentHighlight(
                range=lsp.Range(
                    lsp.Position(line=2, character=0),
                    lsp.Position(line=6, character=0)),
                kind=lsp.DocumentHighlightKind.Text
            )
        ]

    # Code Lens
    @server.feature(lsp.TEXT_DOCUMENT_CODE_LENS)
    def provide_code_lens(ls, params):
        ''' Shows an inline hyperlink that if clicked will trigger a command '''
        return [
            lsp.CodeLens(
                range=lsp.Range(
                    lsp.Position(line=10, character=5),
                    lsp.Position(line=10, character=15)),
                command=lsp.Command(
                    title='My Code Lens',
                    command='logSomething',
                    arguments=['codelens called this', 'some arguments', 'wow']
                ),
            )
        ]

    # Document Symbols
    @server.feature(lsp.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
    def provide_symbols(ls, params):
        ''' Symbols are like function or variable names in a document.

        range, aka fullRange, refers to everything related to the symbol, eg a function body.

        selection_range is a subset of the fullRange, typically just the symbol name.
        '''
        return [
            lsp.DocumentSymbol(
                name='mySymbol1',
                kind=lsp.SymbolKind.Function,
                # function body spans 2 lines
                range=lsp.Range(lsp.Position(line=10, character=0), lsp.Position(line=12, character=0)),
                # function symbol is within the fullRange/range, and spans 3 chars
                selection_range=lsp.Range(lsp.Position(line=10, character=0), lsp.Position(line=10, character=3)),
            ),
            lsp.DocumentSymbol(
                name='mySymbol2',
                kind=lsp.SymbolKind.Function,
                range=lsp.Range(lsp.Position(line=13, character=0), lsp.Position(line=15, character=0)),
                selection_range=lsp.Range(lsp.Position(line=13, character=0), lsp.Position(line=13, character=3)),
            )
        ]

    # Hover
    @server.feature(lsp.TEXT_DOCUMENT_HOVER)
    def hover(ls, params: lsp.TextDocumentPositionParams):
        """Return a hover response."""
        # You'll typically inspect the params to determine the symbol that the
        # user is hovering over and fetch appropriate documentation or other
        # relevant details.
        message = lsp.MarkupContent(
            kind=lsp.MarkupKind.Markdown,  # alt: MarkupKind.PlainText
            value="**This is hover documentation!**\n\nExample content for the hover feature."
        )
        range = lsp.Range(
            start=lsp.Position(line=params.position.line, character=params.position.character),
            end=lsp.Position(line=params.position.line, character=params.position.character + 1)
        )
        return lsp.Hover(contents=message, range=range)

    ##############################
    # Demo LSP Commands

    async def clear_diagnostic_after_delay(uri, delay):
        await asyncio.sleep(delay)
        server.publish_diagnostics(uri, [])

    @server.command("showDiagnostic")
    async def show_diagnostic(ls, args):
        ''' Make a diagnostic of the highlighted region, and then clear it after
        3 seconds. '''
        text_doc = ls.converter.structure(args[0], lsp.TextDocumentIdentifier)
        range = ls.converter.structure(args[1], lsp.Range)

        uri = text_doc.uri
        message = "Sample diagnostic message, will disappear after a few seconds."
        diagnostic = lsp.Diagnostic(
            range=range,
            message=message,
            severity=lsp.DiagnosticSeverity.Warning
        )
        server.publish_diagnostics(uri, [diagnostic])
        asyncio.create_task(clear_diagnostic_after_delay(uri, delay=3))
        return []

    @server.command("showPopup")
    def show_popup(ls, text_doc):
        ''' This shows up in the minibuffer in emacs. '''
        ls.show_message("This is a popup message", lsp.MessageType.Info)

    @server.command("showMsgRequest")
    async def show_msg_request(ls, text_doc):
        params = lsp.ShowMessageRequestParams(
            type=lsp.MessageType.Info,
            message="Is UniteAI the bomb dot com?",
            actions=[
                lsp.MessageActionItem(title="Yes"),
                lsp.MessageActionItem(title="Of course")
            ]
        )
        response = ls.lsp.send_request(lsp.WINDOW_SHOW_MESSAGE_REQUEST, params)
        response = await asyncio.wrap_future(response)
        if response:
            ls.show_message(f"You chose: {response.title}", lsp.MessageType.Info)

    async def send_report(ls):
        # Step 1: Create progress
        token = "sampleToken"
        create_params = lsp.WorkDoneProgressCreateParams(token=token)
        response_future = ls.lsp.send_request(lsp.WINDOW_WORK_DONE_PROGRESS_CREATE, create_params)
        await asyncio.wrap_future(response_future)

        # Step 2: Begin Progress
        progress_begin_params = lsp.ProgressParams(
            token=token,
            value=lsp.WorkDoneProgressBegin(
                title="Analyzing begins",
                cancellable=True,
                message="Starting Message"
            )
        )
        ls.send_notification("$/progress", progress_begin_params)

        # Simulated task with reports
        for percentage in range(0, 101, 10):
            await asyncio.sleep(0.5)  # Simulate some work
            progress_report_params = lsp.ProgressParams(
                token=token,
                value=lsp.WorkDoneProgressReport(
                    percentage=percentage,
                    message=f'Task {percentage}% complete'
                )
            )
            ls.send_notification("$/progress", progress_report_params)

        # Step 4: End Progress
        progress_end_params = lsp.ProgressParams(
            token=token,
            value=lsp.WorkDoneProgressEnd(
                message="Analysis complete"
            )
        )
        ls.send_notification("$/progress", progress_end_params)

    @server.command("reportProgress")
    async def report_progress(ls, args):
        asyncio.create_task(send_report(ls))
        return []

    @server.command("logSomething")
    async def log_something(ls, args):
        '''This is to help debug features that trigger commands, eg code lenses. '''
        ls.show_message(f"You triggered the logSomething command: {str(args)}", lsp.MessageType.Info)
        log.info(f'logSomething log: {str(args)}')
        return []

    # Code Actions
    server.add_code_action(create_action("Example: Show Diagnostic Message", "showDiagnostic"))
    server.add_code_action(create_action("Example: Show Log Message", "showLog"))
    server.add_code_action(create_action("Example: Show Popup Message", "showPopup"))
    server.add_code_action(create_action("Example: Show Message Request", "showMsgRequest"))
    server.add_code_action(create_action("Example: Report Progress", "reportProgress"))

    return server


def post_initialization(config, server):
    # Actor
    server.add_actor(NAME, ExampleActor)

    # Initialize configuration in Actor
    server.tell_actor(NAME, {
        'command': 'set_config',
        'config': config,
    })



'''

Configuration alternatively comes from a config file, or is overridden by a
command line argument.


ORDER OF PRIORITIES

1.) CLI arguments
2.) ./.uniteai.yml
3.) if not found, then read  ~/.uniteai.yml

'''

import argparse
import os
import yaml
import shutil
import logging
import asyncio
from pygls.uris import from_fs_path
from pygls.server import LanguageServer
from lsprotocol.types import (
    INITIALIZE,
    INITIALIZED,
    MessageActionItem,
    MessageType,
    ShowDocumentParams,
    ShowMessageRequestParams,
    WINDOW_SHOW_DOCUMENT,
    WINDOW_SHOW_MESSAGE_REQUEST,
)
import pkg_resources
from uniteai.common import mk_logger
import sys

log = mk_logger('CONFIG', logging.INFO)

CONFIG_PATHS = [
    './.uniteai.yml',
    './.uniteai.yaml',
    os.path.expanduser('~/.uniteai.yml'),
    os.path.expanduser('~/.uniteai.yaml'),
]


def fetch_config():
    '''Entrypoint for fetching configuration.'''
    config_yaml = load_config(CONFIG_PATHS)

    # If no config was found, start the pre-config server
    if config_yaml is None:
        # Try to get config managed via pre-startup language server
        start_config_server()
        config_yaml = load_config(CONFIG_PATHS)  # Reload the config

    if config_yaml is None:
        log.error('Config couldnt be found, nor created. Please copy the example from the github repo and locate it at `~/.uniteai.yml`')
        sys.exit(0)

    # If still no config, just return None
    if config_yaml is None:
        return None, None, None

    # Process CLI arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--stdio', action='store_true', default=True)
    parser.add_argument('--tcp', action='store_true')
    parser.add_argument('--lsp_port', default=config_yaml.get('lsp_port', None))
    parser.add_argument('--modules', default=config_yaml.get('modules', None))

    return parser.parse_args(), config_yaml, parser


def load_config(file_paths=CONFIG_PATHS):
    '''Return first config file that exists.'''
    for file_path in file_paths:
        if os.path.exists(file_path):
            log.info(f'Reading configuration from: {file_path}')
            with open(file_path, 'r') as f:
                return yaml.safe_load(f)
    return None


async def show_document(ls, uri):
    ''' Direct the client to visit a URI. '''
    log.info(f'Visiting document {uri}')
    params = ShowDocumentParams(
        uri=uri,
        external=False,  #  open the document inside the client, not externally (eg web browser)
        take_focus=True,
        selection=None
    )
    try:
        response = ls.lsp.send_request(WINDOW_SHOW_DOCUMENT, params)
    except Exception as e:
        log.error(f'Error showing document {uri}: {e}')
    return asyncio.wrap_future(response)


async def ask_user_with_message_request(ls):
    log.info('Asking user to create new config.')
    params = ShowMessageRequestParams(
        type=MessageType.Info,
        message='No config found for UniteAI. Would you like to create a default config?',
        actions=[
            MessageActionItem(title="Create in current directory"),
            MessageActionItem(title="Create in home directory"),
            MessageActionItem(title="No, I'll do it manually")
        ]
    )
    try:
        response = ls.lsp.send_request(WINDOW_SHOW_MESSAGE_REQUEST, params)
        response = await asyncio.wrap_future(response)
        if response:
            ls.show_message(f"You chose: {response.title}", MessageType.Info)
    except Exception as e:
        log.error(f'Error asking for config creation permission: {e}')
    return response


async def handle_missing_config(server):
    response = await ask_user_with_message_request(server)

    if response:
        title = response.title
        if title == "Create in current directory":
            config_path = './.uniteai.yml'
        elif title == "Create in home directory":
            config_path = os.path.expanduser('~/.uniteai.yml')
        else:
            log.error('Please manually copy and update the file `.uniateai.yml.example` from https://github.com/freckletonj/uniteai.')
            sys.exit(0)

        config_path = os.path.abspath(config_path)
        example_config_path = pkg_resources.resource_filename('uniteai', '.uniteai.yml.example')

        if not os.path.exists(config_path):
            shutil.copyfile(example_config_path, config_path)
            server.show_message(f'Config created at {config_path}', MessageType.Info)

            # visit new file
            response = await show_document(server, from_fs_path(config_path))

            # await asyncio.sleep(1)
            server.show_message(f'Restart UniteAI after editing config!', MessageType.Info)


def start_config_server():
    log.info('Starting the Config Creation Server')
    config_server = LanguageServer('config_creation_server', '0.0.0')

    @config_server.feature(INITIALIZE)
    async def initialize_callback(ls, params):
        log.info(f"Initialize Called")

    @config_server.feature(INITIALIZED)
    async def initialized_callback(ls, params):
        log.info('Config Creation Server successfully initialized')
        # Once the server is initialized, handle the missing config
        try:
            await handle_missing_config(ls)
        except Exception as e:
            log.error(f'Failed making new config: {e}')
        await ls.shutdown()
        await ls.exit()

    config_server.start_io()  # blocks



'''

Document Search.

----------

How to format a request to this feature:

    query:
    docs:
      - title:
        url:
      - title:
        url:

'''


from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Range,
    TextDocumentIdentifier,
    WorkspaceEdit,
)
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from thespian.actors import Actor
import argparse
import logging
import yaml

from uniteai.edit import init_block, cleanup_block, BlockJob
from uniteai.common import extract_range, find_block, mk_logger, get_nested
import uniteai.document.embed as embed
import uniteai.document.download as download
import traceback


##################################################
# Document

START_TAG = ':START_DOCUMENT:'
END_TAG = ':END_DOCUMENT:'
NAME = 'document'
log = mk_logger(NAME, logging.DEBUG)


class DocumentActor(Actor):
    def __init__(self):
        log.debug('ACTOR INIT')
        self.is_running = False
        self.executor = ThreadPoolExecutor(max_workers=3)
        self.current_future = None
        self.should_stop = Event()
        self.tags = [START_TAG, END_TAG]

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        doc = msg.get('doc')
        edits = msg.get('edits')
        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
is_running: {self.is_running}
should_stop: {self.should_stop.is_set()}
current_future: {self.current_future}

EDITS STATE:
job_thread alive: {edits.job_thread.is_alive() if edits and edits.job_thread else "NOT STARTED"}
%%%%%%%%%%''')

        if command == 'set_config':
            config = msg['config']
            self.download_cache = config.download_cache
            self.embedding_cache = config.embedding_cache
            self.db_path = config.db_path
            self.model_name = config.model_name
            self.denoise_window_size = config.denoise_window_size
            self.denoise_poly_order = config.denoise_poly_order
            self.percentile = config.percentile

        if command == 'initialize':
            # import here to speed up startup
            from sentence_transformers import SentenceTransformer

            # Load ML Model
            self.model = SentenceTransformer(self.model_name)

            # Load helper classes
            download.initialize_database(self.db_path)
            self.dl = download.Downloader(self.download_cache)
            self.embedding = embed.Embedding(self.model, self.embedding_cache)
            self.search = embed.Search(self.model,
                                       self.embedding_cache,
                                       self.embedding,
                                       denoise_window_size=1000,
                                       denoise_poly_order=1,
                                       percentile=95,)

        if command == 'start':
            uri = msg.get('uri')
            range = msg.get('range')
            prompt = msg.get('prompt')
            edits = msg.get('edits')

            # check if block already exists
            start_ixs, end_ixs = find_block(START_TAG,
                                            END_TAG,
                                            doc)

            if not (start_ixs and end_ixs):
                init_block(NAME, self.tags, uri, range, edits)

            self.start(uri, range, prompt, edits)

        elif command == 'stop':
            self.stop()

    def start(self, uri, range, prompt, edits):
        if self.is_running:
            log.info('WARN: ON_START_BUT_RUNNING')
            return
        log.debug('ACTOR START')

        self.is_running = True
        self.should_stop.clear()

        def f(uri_, prompt_, should_stop_, edits_):
            ''' Compose the streaming fn with some cleanup. '''
            log.debug('START: DOCUMENT_STREAM_FN')
            try:
                args = yaml.safe_load(prompt_)

                # Format of documents to load
                #
                #     query:
                #     docs:
                #       - title:
                #         url:
                #       - title:
                #         url:

                query = args['query']
                docs = args['docs']

                # TODO this is ugly
                titles_urls = []
                for d in docs:
                    titles_urls.append((d['title'] if 'title' in d else None,
                                        d['url']))
                download.save_docs(titles_urls, self.dl, self.db_path)

                search_results = embed.search_helper(
                    self.search,
                    self.db_path,
                    titles_urls,
                    query,
                    window_size=2000,
                    stride=50,
                    top_n=3,
                    visualize=False)

                output = output_template(prompt_, search_results)
                # emacs lsp-mode doesn't like unicode NULLs
                output = output.replace('\u0000', '')
                job = BlockJob(
                    uri=uri,
                    start_tag=START_TAG,
                    end_tag=END_TAG,
                    text=f'\n{output}\n',
                    strict=True,
                )
                edits.add_job(NAME, job)

            except Exception as e:
                log.error(f'Error: Document, {e}')
                error_info = traceback.format_exc()
                log.error(error_info)

            # Cleanup
            log.debug('CLEANING UP')
            cleanup_block(NAME, self.tags, uri_, edits_)
            self.is_running = False
            self.current_future = None
            self.should_stop.clear()

        self.current_future = self.executor.submit(
            f, uri, prompt, self.should_stop, edits
        )
        log.debug('START CAN RETURN')

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running:
            log.info('WARN: ON_STOP_BUT_STOPPED')

        self.should_stop.set()

        if self.current_future:
            self.current_future.result()  # block, wait to finish
            self.current_future = None
        log.debug('FINALLY STOPPED')


def output_template(prompt, search_results):
    out = ''
    for url, title, results in search_results:
        for i, result in enumerate(results):
            out += f'''
--------------------------------------------------
TITLE: {title}
URL: {url}
RESULT {i+1} / {len(results)}:

{result}

'''
        out += '\n\n'
    return out


##################################################
# Document

def code_action_document(params: CodeActionParams):
    '''Trigger a GPT Autocompletion response. A code action calls a command,
    which is set up below to `tell` the actor to start streaming a response.'''
    text_document = params.text_document
    range = params.range
    return CodeAction(
        title='Document',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='Document',
            command='command.document',
            # Note: these arguments get jsonified, not passed as python objs
            arguments=[text_document, range]
        )
    )


##################################################
# Setup

def configure(config_yaml):
    parser = argparse.ArgumentParser()
    parser.add_argument('--download_cache', default=get_nested(config_yaml, ['document', 'download_cache']))
    parser.add_argument('--embedding_cache', default=get_nested(config_yaml, ['document', 'embedding_cache']))
    parser.add_argument('--db_path', default=get_nested(config_yaml, ['document', 'db_path']))
    parser.add_argument('--model_name', default=get_nested(config_yaml, ['document', 'model_name']))
    parser.add_argument('--denoise_window_size', default=get_nested(config_yaml, ['document', 'denoise_window_size']))
    parser.add_argument('--denoise_poly_order', default=get_nested(config_yaml, ['document', 'denoise_poly_order']))
    parser.add_argument('--percentile', default=get_nested(config_yaml, ['document', 'percentile']))

    # bc this is only concerned with this module's params, do not error if
    # extra params are sent via cli.
    args, _ = parser.parse_known_args()
    return args


def initialize(config, server):
    # CodeActions
    server.add_code_action(
        lambda params:
        code_action_document(params))

    # Modify Server
    @server.thread()
    @server.command('command.document')
    def document(ls, args):
        if len(args) != 2:
            log.error(f'command.document: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        range = ls.converter.structure(args[1], Range)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Extract the highlighted region
        prompt = extract_range(doc_source, range)

        # Send a message to start the stream
        actor_args = {
            'command': 'start',
            'uri': uri,
            'range': range,
            'prompt': prompt,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)

        # Return null-edit immediately (the rest will stream)
        return WorkspaceEdit()


def post_initialization(config, server):
    server.add_actor(NAME, DocumentActor)
    server.tell_actor(NAME, {
        'command': 'set_config',
        'config': config
    })
    server.tell_actor(NAME, {
        'command': 'initialize'
    })



'''

File Download

'''

from bs4 import BeautifulSoup
import os
import requests
import re
from concurrent.futures import ThreadPoolExecutor
import sqlite3
from git import Repo
from typing import List, Tuple, Union
from pathlib import Path
from io import BytesIO
from youtube_transcript_api import YouTubeTranscriptApi
from uniteai.common import read_unicode
from uniteai.common import mk_logger
import logging


# TQDM can be useful for debugging, but prints to stderr which makes VSCode
# flip out.
DEBUG = False
if DEBUG:
    from tqdm import tqdm
else:
    def tqdm(x, *args, **kwargs):
        return x

FILETYPES = [
    '.pdf',
    '.py',  '.pyc', '.ipynb',
    '.txt', '.md', '.org',
    '.java',
    '.c', '.h', '.cpp',
    '.js', '.html', '.css', '.ts', '.jsx',
    '.hs',
]


log = mk_logger('document-download', logging.DEBUG)


##################################################
# Youtube

def extract_video_id(url):
    regex_patterns = [
        r"youtube\.com/watch\?v=([a-zA-Z0-9_-]{11})",      # Standard URL
        r"youtu\.be/([a-zA-Z0-9_-]{11})",                  # Shortened URL
        r"youtube\.com/embed/([a-zA-Z0-9_-]{11})",         # Embed URL
        r"youtube\.com/v/([a-zA-Z0-9_-]{11})",             # V URL
        # URL with timestamp
        r"youtube\.com/watch\?v=([a-zA-Z0-9_-]{11})&.*#t=(\d+)m(\d+)s",
    ]
    for pattern in regex_patterns:
        match = re.search(pattern, url)
        if match:
            return match.group(1)
    return None


headers = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.0.0 Safari/537.36'
}


def get_video_title(url: str):
    ''' Try to get the title of a Youtube video. '''
    response = requests.get(url, headers=headers)
    soup = BeautifulSoup(response.text, 'html.parser')
    return soup.find('meta', property="og:title")["content"]


def get_transcript(url):
    ''' Try to get the transcript of a Youtube video. '''
    video_id = extract_video_id(url)
    if video_id is None:
        log.error("Could not extract video ID from URL")
        return None

    try:
        transcript = YouTubeTranscriptApi.get_transcript(video_id)
        full_transcript = ""
        for line in transcript:
            full_transcript += " " + line['text']
        return full_transcript
    except Exception as e:
        log.error("An error occurred:", e)
        return None


##################################################
# Download

def sanitize(filename):
    ''' Sanitize strings for becoming filenames. '''
    # Replace any characters not allowed in filenames with underscores
    sanitized_filename = re.sub(r'[^\w\-_.]', '_', filename)

    # Remove multiple consecutive underscores
    sanitized_filename = re.sub(r'__+', '_', sanitized_filename)

    # Remove leading and trailing underscores
    sanitized_filename = sanitized_filename.strip('_')

    # Limit the filename length to 255 characters (a common limitation on some systems)
    max_filename_length = 255
    sanitized_filename = sanitized_filename[:max_filename_length]

    return sanitized_filename


def normalize_local_uri(uri):
    if uri.startswith('file://'):
        uri = uri[len('file://'):]
    uri = os.path.expanduser(uri)
    return uri


def is_local_uri(uri):
    uri = normalize_local_uri(uri)
    if os.path.exists(uri):
        return True
    else:
        return False


def read_from_disk(path: Path) -> List[Tuple[Path, BytesIO]]:
    files = []

    # If path is a directory
    if path.is_dir():
        for file_path in path.iterdir():
            if file_path.suffix in FILETYPES:
                with open(file_path, 'rb') as fp:
                    buf = BytesIO(fp.read())
                    files.append((file_path, buf))

    # If path is a file
    elif path.is_file() and path.suffix in FILETYPES:
        with open(path, 'rb') as fp:
            buf = BytesIO(fp.read())
            files.append((path, buf))

    # Path is neither a file nor a directory (or it's an unsupported filetype)
    else:
        log.warning(f"Unsupported path or filetype: {path}")

    return files


class Downloader:
    '''
    1. Download and cache a url to disk
    2. Fetch it, and convert nicely to unicode
    '''

    def __init__(self, cache_dir: str):
        self.cache_dir = Path(cache_dir)

    def _get_extension(self, content_type: str) -> str:
        mapping = {'text/html': '.html',
                   'application/json': '.json',
                   'text/plain': '.txt',
                   'application/pdf': '.pdf'
                   }
        # a lookup won't work in cases where, eg,
        #   Content-Type: text/plain; charset=utf-8
        for k, ext in mapping.items():
            if content_type in k:
                return ext
        return '.txt'

    def _preprocess_url(self, url: str):
        ''' Some URLs deserve special treatment. '''
        if 'arxiv.org' in url:
            url = url.replace('abs', 'pdf')
        return url

    def fetch(self, title: Union[str, None], uri: str) -> List[BytesIO]:
        ''' Fetches bytes of a file (either from disk or online) based on the provided URI. '''
        if is_local_uri(uri):
            return self.fetch_offline(uri)
        else:
            return self.fetch_online(title, uri)

    def fetch_offline(self, uri: str) -> List[BytesIO]:
        ''' Fetches bytes of a file from the local disk based on the provided URI. '''
        uri = normalize_local_uri(uri)
        dir_path = Path(uri)

        if not dir_path.exists():
            log.error(f'Local path does not exist: {uri}')
            return []

        return read_from_disk(dir_path)

    def fetch_online(self, title: Union[str, None], url: str) -> List[BytesIO]:
        ''' Fetches the bytes of a file, reads from cache if it already exists,
        or downloads otherwise.

        Notice it returns a List because in cases like a git repo, there are
        many files represented by one URL.'''
        # Preprocess
        url = self._preprocess_url(url)

        url_sane = sanitize(url)  # instead of hashing
        dir_path = self.cache_dir / url_sane

        # Save if not cached
        if not dir_path.exists():
            log.info(f'URL hasnt been cached, fetching')
            # Github
            if 'github.com' in url:
                log.info(f'Downloading Github URL: {url}')
                os.makedirs(dir_path, exist_ok=True)
                Repo.clone_from(url, dir_path)

            # Youtube
            elif 'youtube.com' in url or 'youtu.be' in url:
                log.info(f'Downloading Youtube URL: {url}')
                video_id = extract_video_id(url)
                if video_id is None:
                    log.error("Could not extract video ID from URL")
                else:
                    try:
                        title = get_video_title(url)
                        if title is not None:
                            title = sanitize(title)
                        else:
                            title = video_id

                        transcript = YouTubeTranscriptApi.get_transcript(video_id)
                        full_transcript = ""
                        for line in transcript:
                            full_transcript += " " + line['text']
                        file_path = dir_path / f'{title}.txt'

                        os.makedirs(dir_path, exist_ok=True)
                        with open(file_path, 'w') as fp:
                            fp.write(full_transcript)
                    except Exception as e:
                        log.error("An error occurred:", e)

            # Other
            else:
                try:
                    response = requests.get(url, headers=headers)
                except requests.exceptions.SSLError as e:
                    log.error(f'SSLError: {e}')
                    # response = requests.get(url, headers=headers, verify=False)
                    return []

                if response.status_code == 200:
                    file_ext = self._get_extension(response.headers.get('Content-Type', ''))

                    # HTML
                    if file_ext == '.html':
                        log.info(f'Downloading URL as HTML: {url}')
                        content = response.text
                        soup = BeautifulSoup(content, 'html.parser')

                        # infer title
                        if title is None:
                            # Try to find HTML title in meta tags
                            meta_title = soup.find('meta', property="og:title")
                            if meta_title is None:
                                meta_title = soup.find('meta', property="title")
                            if meta_title is not None:
                                meta_title = meta_title["content"]

                            # Try to find HTML title in `title`
                            if meta_title is None:
                                meta_title = soup.find('title')
                            if meta_title is not None:
                                meta_title = meta_title.getText()

                            # If found a meta_title, keep it
                            if meta_title is not None:
                                title = meta_title

                        # backup title
                        if title is None:
                            title = url_sane

                        file_path = dir_path / f'{title}{file_ext}'
                        os.makedirs(dir_path, exist_ok=True)
                        with open(file_path, 'w') as fp:
                            log.debug(f'writing: {file_path}')
                            fp.write(soup.get_text(separator=' '))

                    # PDF
                    elif file_ext == '.pdf':
                        log.info(f'Downloading URL as PDF: {url}')
                        if title is None:
                            title = url_sane
                        file_path = dir_path / f'{title}{file_ext}'
                        os.makedirs(dir_path, exist_ok=True)
                        with open(file_path, 'wb') as fp:  # note the 'wb' here
                            log.debug(f'writing: {file_path}')
                            fp.write(response.content)  # note the use of content instead of text here

                    # OTHER
                    else:
                        log.info(f'Downloading URL as "Other": {url}')
                        content = response.text
                        if title is None:
                            title = url_sane
                        file_path = dir_path / f'{title}{file_ext}'
                        os.makedirs(dir_path, exist_ok=True)
                        with open(file_path, 'w') as fp:
                            log.debug(f'writing: {file_path}')
                            fp.write(content)
                else:
                    log.error(f'NOT FOUND: {url}, {response.status_code}, {response}')

        files = read_from_disk(dir_path)
        return files

    def fetch_utf8(self, title: Union[str, None], url: str) -> List[Tuple[str, str]]:
        ''' Fetches and converts files to unicode. Reads from cache if it
        already exists, or downloads otherwise. '''
        files = self.fetch(title, url)
        contents = []
        for file_path, buf in files:
            contents.append((file_path, read_unicode(file_path, buf)))
        return contents


##################################################
# Database

migrations = ['''
CREATE TABLE IF NOT EXISTS resources (
    id INTEGER PRIMARY KEY,
    title TEXT,
    url TEXT,
    file_path TEXT,
    content TEXT
);''',

'''
CREATE UNIQUE INDEX IF NOT EXISTS resources_unique_url_file_path ON resources(url, file_path);
''',
]


def initialize_database(db_path):
    connection = sqlite3.connect(db_path)
    cursor = connection.cursor()
    for mig in migrations:
        cursor.execute(mig)
        connection.commit()


def insert_to_database(connection, url: str, title: str, file_path: str, content: str):
    cursor = connection.cursor()
    cursor.execute("""
INSERT OR IGNORE INTO resources (url, title, file_path, content)
VALUES (?, ?, ?, ?)
    """, (url, title, file_path, content))
    connection.commit()


##################################################
# Go

def save_url(title: Union[str, None], url: str, dl: Downloader, db_path):
    files = dl.fetch_utf8(title, url)
    with sqlite3.connect(db_path) as conn:
        for path, content in files:
            if title is None:
                title = path.stem
            # print(f'{title}, {path.stem}, {url[:20]}, {str(path)[-20:]}, {content[:100]}')
            insert_to_database(conn, url, title, str(path), content)
        return files


def save_docs(references, dl, db_path):
    with ThreadPoolExecutor(max_workers=3) as executor:
        futures = []
        for title, url in references:
            if url is None:
                log.error("No URL provided")
                continue

            futures.append(executor.submit(save_url, title, url, dl, db_path))

        # Wait for all downloads to finish
        out = []
        for future in tqdm(futures, desc="Fetching"):
            files = future.result()
            out += files
        return out



'''

Launch an LLM locally, and serve it.


----------
RUN:
    uvicorn llm_server:app --port 8000


----------

NOTE: Multiprocessing could potentially be nice instead of Threads,
     BUT Threads is working great, AND:

I've Struggled to get MP to work.

You need to place at top of file:
    mp.set_start_method('spawn')
CUDA requires this, but Uvicorn already forces it to be `fork`.

Perhaps Uvicorn has some nice spawn-friendly process ability?

If you can get that to work, swap for the Thread lines in the llm_streaming
endpoint:
    q = mp.Manager().Queue()
    mp.Process(
        target=local_llm_stream_,
        args=(
            request,
            q,
            local_llm_stop_event
        )).start()

'''


from fastapi import FastAPI, HTTPException
from fastapi.responses import StreamingResponse, Response
from pydantic import BaseModel
from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    TextIteratorStreamer,
    T5Tokenizer,
    T5ForConditionalGeneration,
)
from transformers.generation import StoppingCriteriaList
from typing import List
import threading
import torch
import multiprocessing as mp
from uniteai.common import get_nested
from uniteai.config import load_config
import uvicorn


device = 'cuda' if torch.cuda.is_available() else 'cpu'

##################################################
# `transformers`


def load_model(args):
    name_or_path = args['model_name_or_path']

    # T5
    if 't5' in name_or_path:
        tokenizer = T5Tokenizer.from_pretrained(name_or_path)
        model = T5ForConditionalGeneration.from_pretrained(name_or_path, device_map='auto')
        return tokenizer, model

    # AutoModelForCausalLM (should support many models)
    else:
        tokenizer = AutoTokenizer.from_pretrained(
            name_or_path,
        )
        revision = {'revision': args['model_commit']} if 'model_commit' in args else {}
        device_map = {'device_map': args['device_map']} if 'device_map' in args else {}
        load_in_8bit = {'load_in_8bit': args['load_in_8bit']} if 'load_in_8bit' in args else {}
        load_in_4bit = {'load_in_4bit': args['load_in_4bit']} if 'load_in_4bit' in args else {}
        trust_remote_code = {'trust_remote_code': args['trust_remote_code']} if 'trust_remote_code' in args else {}

        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=name_or_path,
            **trust_remote_code,  # needed by eg Falcon, and MosaicML
            **revision,
            **device_map,
            **load_in_8bit,
            **load_in_4bit,
        )
        return tokenizer, model


##################################################
# Initialization

config = load_config()
args = config['local_llm']


##################################################
# Types

class AutocompleteRequest(BaseModel):
    text: str
    max_length: int = 200
    do_sample: bool = True
    top_k: int = 10
    num_return_sequences: int = 1


class AutocompleteResponse(BaseModel):
    generated_text: str


app = FastAPI()


@app.on_event("startup")
def initialize_model():
    global tokenizer, model
    tokenizer, model = load_model(args)


##################################################
# Local LLM

# for early termination of streaming
local_llm_stop_event = mp.Event()

# when streaming, a chunk must end in newline
STREAM_TOK = '\n'


def local_llm_stream_(request, streamer, local_llm_stop_event):
    def custom_stopping_criteria(local_llm_stop_event):
        def f(input_ids: torch.LongTensor,
              score: torch.FloatTensor,
              **kwargs) -> bool:
            return local_llm_stop_event.is_set()
        return f
    stopping_criteria = StoppingCriteriaList([
        custom_stopping_criteria(local_llm_stop_event)
    ])

    toks = tokenizer([request.text], return_tensors='pt').to(device)
    generation_kwargs = dict(
        inputs=toks.input_ids.to(device),
        attention_mask=toks.attention_mask,
        streamer=streamer,
        max_length=request.max_length,
        do_sample=request.do_sample,
        top_k=request.top_k,
        num_return_sequences=request.num_return_sequences,
        eos_token_id=tokenizer.eos_token_id,
        pad_token_id=tokenizer.eos_token_id,  # for open-end generation
        stopping_criteria=stopping_criteria,
        # important for when model is accessed from other threads
        #   https://github.com/h2oai/h2ogpt/pull/297/files
        use_cache=False,
    )
    try:
        model.generate(**generation_kwargs)  # blocks
    except RuntimeError as e:
        # NOTE: Falcon randomly produces errors like:
        #       mat1 and mat2 shapes cannot be multiplied
        streamer.on_finalized_text(f'\n<LLM SERVER ERROR: {e}>',
                                   stream_end=True)
    print('DONE GENERATING')


def stream_results(streamer):
    for x in streamer:
        # Stream response, and stop early if requested
        if not local_llm_stop_event.is_set():
            print(f'yield: {x}')
            yield (
                AutocompleteResponse(generated_text=x).json() + STREAM_TOK
            ).encode('utf-8')
        else:
            print('STOP WAS SET')
            break


@app.post("/local_llm_stream", response_model=List[AutocompleteResponse])
def local_llm_stream(request: AutocompleteRequest):
    ''' Stream the response as the model generates it. '''
    local_llm_stop_event.clear()
    streamer = TextIteratorStreamer(
        tokenizer,
        skip_special_tokens=True  # eg <|endoftext|>
    )
    threading.Thread(target=local_llm_stream_,
                     args=(
                         request,
                         streamer,
                         local_llm_stop_event
                     )).start()

    try:
        return StreamingResponse(
            stream_results(streamer),
            media_type="application/json"
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/local_llm_stream_stop")
def local_llm_stream_stop():
    ''' End ongoing inferrence '''
    global local_llm_stop_event
    local_llm_stop_event.set()
    return Response(status_code=200)


def main():
    uvicorn.run("uniteai.llm_server:app",
                host=get_nested(config, ['local_llm', 'host']),
                port=get_nested(config, ['local_llm', 'port']))


if __name__ == "__main__":
    main()



'''

A locally running LLM.

'''

from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    Command,
    Range,
    TextDocumentIdentifier,
)
import requests
import json
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from thespian.actors import Actor
import logging
import argparse

from uniteai.edit import init_block, cleanup_block, BlockJob
from uniteai.common import extract_range, find_block, mk_logger, get_nested


START_TAG = ':START_LOCAL:'
END_TAG = ':END_LOCAL:'
NAME = 'local_llm'
log = mk_logger(NAME, logging.WARN)


class LocalLLMActor(Actor):
    def __init__(self):
        log.debug('ACTOR INIT')
        self.is_running = False
        self.executor = ThreadPoolExecutor(max_workers=3)
        self.current_future = None
        self.should_stop = Event()
        self.tags = [START_TAG, END_TAG]

    def receiveMessage(self, msg, sender):
        command = msg.get('command')
        doc = msg.get('doc')

        edits = msg.get('edits')
        log.debug(f'''
%%%%%%%%%%
ACTOR RECV: {msg["command"]}
ACTOR STATE:
is_running: {self.is_running}
should stop: {self.should_stop.is_set()}
current_future: {self.current_future}

EDITS STATE:
job_thread alive: {edits.job_thread.is_alive() if edits and edits.job_thread else "NOT STARTED"}
%%%%%%%%%%''')

        ##########
        # Start
        if command == 'start':
            uri = msg.get('uri')
            range = msg.get('range')
            prompt = msg.get('prompt')
            edits = msg.get('edits')

            # check if block already exists
            start_ixs, end_ixs = find_block(START_TAG,
                                            END_TAG,
                                            doc)

            if not (start_ixs and end_ixs):
                init_block(NAME, self.tags, uri, range, edits)

            self.start(uri, range, prompt, edits)

        ##########
        # Stop
        elif command == 'stop':
            self.stop()

        ##########
        # Config
        elif command == 'set_config':
            self.max_length = msg.get('local_llm_max_length')
            self.top_k = msg.get('local_llm_top_k')
            self.port = msg.get('local_llm_port')
            self.host = msg.get('local_llm_host')
            self.llm_uri = f'http://{self.host}:{self.port}'

    def start(self, uri, range, prompt, edits):
        if self.is_running:
            log.info('WARN: ON_START_BUT_RUNNING')
            return
        log.debug('ACTOR START')

        self.is_running = True
        self.should_stop.clear()

        self.current_future = self.executor.submit(
            self.local_llm_stream_fn, uri, prompt, self.should_stop, edits
        )
        log.debug('START CAN RETURN')

    def stop(self):
        log.debug('ACTOR STOP')
        if not self.is_running:
            log.info('WARN: ON_STOP_BUT_STOPPED')

        self.should_stop.set()

        if self.current_future:
            self.current_future.result()  # block, wait to finish
            self.current_future = None
        log.debug('FINALLY STOPPED')

    def local_llm_stream_fn(self, uri, prompt, stop_event, edits):
        log.debug('START: LOCAL_LLM_STREAM_FN')
        try:
            request_data = {
                "text": prompt,
                "max_length": self.max_length,
                "do_sample": True,
                "top_k": self.top_k,
                "num_return_sequences": 1
            }
            # Stream response from LLM Server
            response = requests.post(f"{self.llm_uri}/local_llm_stream",
                                     json=request_data,
                                     stream=True)
            if response.status_code != 200:
                raise Exception(
                    f"POST request to {self.llm_uri} failed with status code "
                    f"{response.status_code}"
                )

            # Stream the results to LSP Client
            running_text = ''
            for line in response.iter_lines():
                # For breaking out early
                if stop_event.is_set():
                    log.debug('STREAM_FN received STOP EVENT')
                    break
                response_data = json.loads(line)
                new_text = response_data["generated_text"]
                log.debug(f'NEW: {new_text}')
                # ignore empty strings
                if len(new_text) == 0:
                    continue

                running_text += new_text
                job = BlockJob(
                    uri=uri,
                    start_tag=START_TAG,
                    end_tag=END_TAG,
                    text=f'\n{running_text}\n',
                    strict=False,
                )
                edits.add_job(NAME, job)

            # Streaming is done, and those added jobs were all
            # non-strict. Let's make sure to have one final strict
            # job. Streaming jobs are ok to be dropped, but we need to make
            # sure it does finalize, eg before a strict delete-tags job is
            # added.
            job = BlockJob(
                uri=uri,
                start_tag=START_TAG,
                end_tag=END_TAG,
                text=f'\n{running_text}\n',
                strict=True,
            )
            edits.add_job(NAME, job)

        except requests.exceptions.ConnectionError as e:
            # LLM Server isn't started
            msg = f'\nERROR, first start the local LLM Server as a separate process: `uniteai_llm`.\n\nERROR MSG: {e}'
            log.error(msg)
            job = BlockJob(
                uri=uri,
                start_tag=START_TAG,
                end_tag=END_TAG,
                text=msg,
                strict=True,
            )
            edits.add_job(NAME, job)

        except Exception as e:
            log.error(f'Error: Local LLM, {e}')

        # Cleanup
        log.debug('CLEANING UP')
        cleanup_block(NAME, self.tags, uri, edits)
        self.is_running = False
        self.current_future = None
        self.should_stop.clear()


def code_action_local_llm(params: CodeActionParams):
    text_document = params.text_document
    range = params.range
    return CodeAction(
        title='Local LLM',
        kind=CodeActionKind.Refactor,
        command=Command(
            title='Local LLM',
            command='command.localLlmStream',
            # Note: these arguments get jsonified, not passed directly
            arguments=[text_document, range]
        )
    )


##################################################
# External API

def configure(config_yaml):
    parser = argparse.ArgumentParser()

    parser.add_argument('--local_llm_model_path', default=get_nested(config_yaml, ['local_llm', 'model_path']))
    parser.add_argument('--local_llm_model_commit', default=get_nested(config_yaml, ['local_llm', 'model_commit']))
    parser.add_argument('--local_llm_max_length', default=get_nested(config_yaml, ['local_llm', 'max_length']))
    parser.add_argument('--local_llm_top_k', default=get_nested(config_yaml, ['local_llm', 'top_k']))
    parser.add_argument('--local_llm_port', default=get_nested(config_yaml, ['local_llm', 'port']))
    parser.add_argument('--local_llm_host', default=get_nested(config_yaml, ['local_llm', 'host']))

    # bc this is only concerned with local llm params, do not error if extra
    # params are sent via cli.
    args, _ = parser.parse_known_args()
    return args


def initialize(config, server):
    # Actor
    server.add_actor(NAME, LocalLLMActor)
    server.tell_actor(NAME, {
        'command': 'set_config',
        **vars(config)  # argparse.Namespace -> dict
    })

    # CodeActions
    server.add_code_action(code_action_local_llm)

    @server.thread()
    @server.command('command.localLlmStream')
    def local_llm_stream(ls, args):
        if len(args) != 2:
            log.error(f'command.localLlmStream: Wrong arguments, received: {args}')
        text_document = ls.converter.structure(args[0], TextDocumentIdentifier)
        range = ls.converter.structure(args[1], Range)
        uri = text_document.uri
        doc = ls.workspace.get_document(uri)
        doc_source = doc.source

        # Extract the highlighted region
        prompt = extract_range(doc_source, range)

        # Send a message to start the stream
        actor_args = {
            'command': 'start',
            'uri': uri,
            'range': range,
            'prompt': prompt,
            'edits': ls.edits,
            'doc': doc_source,
        }
        ls.tell_actor(NAME, actor_args)
